from typing import List, Optional

from rdflib import Literal, URIRef, XSD
import rdflib.plugins.sparql.processor

from rdf_cty_ccy.graph import graph_query
from rdf_cty_ccy.common import fn, monad
from rdf_cty_ccy.model import model


class Filter:
    WithCurrency = 'withCurrency'


def by_country_code(code: str, filters: List[Optional[Filter]] = []) -> Optional[model.Country]:
    """
    Takes an ISO3166 Alpha-3 country code and returns a single country/currency result.
    If there are no results, None is returned.  If a code as string was not provided, None is returned.

    :param code: A single ISO3166 3 Alpha country code
    :param filters: An Optional List of filters directing what should be returned; from the Filter enum
    :return: Country object with associated Currency
    """
    if not isinstance(code, str):
        return None
    lit = convert_literal(code)
    if lit.is_left():
        return None
    return build(model.Country(), find_by_code(cty_tag_maximal_query, lit.value), filters)


def by_country_uri(uri: str, filters: List[Optional[Filter]] = []) -> Optional[model.Country]:
    if not isinstance(uri, str):
        return None
    uriref = convert_uri(uri)
    if uriref.is_left():
        return None
    return build(model.Country(country_uri=uriref.value), find_by_uri(cty_uri_maximal_query, uriref.value), filters)


@monad.monadic_try()
def convert_literal(lit):
    return Literal(lit, datatype=XSD.string)


@monad.monadic_try()
def convert_uri(uri):
    return URIRef(uri)


def find_by_code(query_fn, lit_code):
    return [r for r in graph_query.sparql(query_fn(lit_code))]


def find_by_uri(query_fn, uri):
    return [r for r in graph_query.sparql(query_fn(uri))]


def build(country: model.Country,
          result: rdflib.query.ResultRow,
          filters: List[Optional[Filter]]) -> Optional[model.Country]:
    if not result:
        return None

    cty = fn.first(result)

    if not country.country_uri:
        country.country_uri = cty.cty

    country.identifies = cty.cty_identifies
    country.code_type = cty.codeset
    country.label = cty.cty_label

    if Filter.WithCurrency in filters:
        if len(result) == 1:
            ccy = build_ccy(cty)
            country.currency = ccy
            country.currencies = [ccy]
        else:
            country.currencies = list(map(build_ccy, result))
    return country


def build_ccy(result) -> model.Currency:
    return model.Currency(currency_uri=result.ccy_code_uri,
                          identifies=result.ccy_identifies,
                          label=result.ccy_label)


def cty_tag_maximal_query(cty_tag: Literal) -> rdflib.plugins.sparql.processor.SPARQLResult:
    return """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX lcc-lr: <https://www.omg.org/spec/LCC/Languages/LanguageRepresentation/>
    PREFIX lcc-cr: <https://www.omg.org/spec/LCC/Countries/CountryRepresentation/>

    SELECT ?cty ?cty_identifies ?cty_label ?codeset ?ccy_identifies ?ccy_code_uri ?ccy_label
    WHERE {{
      ?cty lcc-lr:hasTag {cty_tag} ;
           rdfs:label ?cty_label ; 
           lcc-lr:isMemberOf ?codeset ;
           lcc-lr:identifies ?cty_identifies .
      ?ccy_identifies lcc-cr:isUsedBy ?cty_identifies .
      ?ccy_code_uri lcc-lr:identifies ?ccy_identifies ;
                   rdfs:label ?ccy_label  
    }}
    """.format(cty_tag=cty_tag.n3())


def cty_uri_maximal_query(cty_uri: URIRef) -> rdflib.plugins.sparql.processor.SPARQLResult:
    return """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
    PREFIX lcc-lr: <https://www.omg.org/spec/LCC/Languages/LanguageRepresentation/>
    PREFIX lcc-cr: <https://www.omg.org/spec/LCC/Countries/CountryRepresentation/>

    SELECT ?cty ?cty_identifies ?codeset ?cty_label ?ccy_identifies ?ccy_code_uri ?ccy_label
    WHERE {{
      {cty_uri} rdfs:label ?cty_label ; 
           lcc-lr:isMemberOf ?codeset ;
           lcc-lr:identifies ?cty_identifies .
      ?ccy_identifies lcc-cr:isUsedBy ?cty_identifies .
      ?ccy_code_uri lcc-lr:identifies ?ccy_identifies ;
                   rdfs:label ?ccy_label  
    }}
    """.format(cty_uri=cty_uri.n3())



from rdflib import Graph
from simple_memory_cache import GLOBAL_CACHE
from pathlib import Path
import pickle

base_path = Path(__file__).parent

from . import binding

graph_cache = GLOBAL_CACHE.MemoryCachedVar('graph_cache')

from_pickle = True

currency_code_triples = (base_path / "../rdfdata/ISO3166-1-CountryCodes.ttl").resolve()
full_graph_rdf_pickle = (base_path / "../rdfdata/cty-ccy-rdf.p").resolve()
country_code_triples = (base_path / "../rdfdata/ISO4217-CurrencyCodes.ttl").resolve()
sfo_currency_extensions = (base_path / '../rdfdata/sfo-currency-extensions.ttl').resolve()

def graph():
    return graph_cache.get()

def invalidate_cache():
    graph_cache.invalidate()
    pass

@graph_cache.on_first_access
def build_graph():
    if from_pickle:
        return load_graph_from_pickle()

    return load_from_ttl()

def load_from_ttl():
    g = initgraph()
    g.parse(currency_code_triples)
    g.parse(country_code_triples)
    g.parse(sfo_currency_extensions)
    return g


def load_graph_from_pickle():
    f = open(full_graph_rdf_pickle, 'rb')
    g = pickle.load(f)
    f.close()
    return g

def initgraph() -> Graph:
    return binding.bind(rdf_graph())

def rdf_graph():
    return Graph()


import argparse
import requests
import os
import pickle

from rdf_cty_ccy.graph import graph

ccy_url = "https://spec.edmcouncil.org/fibo/ontology/master/2022Q1/FND/Accounting/ISO4217-CurrencyCodes/"
cty_url = "https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/"

ccy_temp = "rdf_cty_ccy/rdfdata/ccy.rdf"
cty_temp = "rdf_cty_ccy/rdfdata/cty.rdf"

cty_ttl_loc = "rdf_cty_ccy/rdfdata/ISO3166-1-CountryCodes.ttl"
ccy_ttl_loc = "rdf_cty_ccy/rdfdata/ISO4217-CurrencyCodes.ttl"
ccy_ext_ttl_loc = "rdf_cty_ccy/rdfdata/sfo-currency-extensions.ttl"

ccy_rdf_pickle = "rdf_cty_ccy/rdfdata/ISO3166-1-CountryCodes.p"
cty_rdf_pickle = "rdf_cty_ccy/rdfdata/ISO4217-CurrencyCodes.p"

full_graph_rdf_pickle = "rdf_cty_ccy/rdfdata/cty-ccy-rdf.p"

def builder(args=None):
    """
    Examples:
          poetry run build_onto --loc rdf_cty_ccy/rdfdata/
    """

    parser = argparse.ArgumentParser(description='Downloads CCY and CTY ontologies and builds as TTL')
    parser.add_argument('--loc', type=str, nargs='?', default="rdf_cty_ccy/rdfdata/",
                        help='Location for the TTL files')

    if not args:
        args = parser.parse_args()

    ccy_result = create_ttl(url=ccy_url, temp=ccy_temp, ttl_out=ccy_ttl_loc, pickle_out=ccy_rdf_pickle)
    cty_result = create_ttl(url=cty_url, temp=cty_temp, ttl_out=cty_ttl_loc, pickle_out=cty_rdf_pickle)
    create_pickled_rdf(pickle_out=full_graph_rdf_pickle)
    pass


def create_ttl(url, temp, ttl_out, pickle_out):
    print("Building: {url} to {ttl}".format(url=url, ttl=ttl_out))
    result = requests.get(url)
    with open(temp, 'w') as f:
        f.write(result.text)

    g = graph.rdf_graph()
    g.parse(temp)
    write_to_ttl(g, format="turtle", file=ttl_out)
    os.remove(temp)
    print("Result: OK")
    return "OK"

def write_to_ttl(g, format="turtle", file=None):
    if format == "turtle":
        txt = g.serialize(format=format)
    else:
        txt = g.serialize(format="json-ld", indent=4)
    if file:
        with open(file, 'w') as f:
            f.write(txt)

def create_pickled_rdf(pickle_out):
    print("Creating Pickled RDF: to {pickle}".format(pickle=pickle_out))
    g = graph.load_from_ttl()
    write_to_pickle(g, pickle_out)
    print("Result: OK")
    pass

def write_to_pickle(g, file):
    f = open(file, 'wb')
    pickle.dump(g, f)
    f.close()
    pass



from typing import List, Tuple, Optional, Union

import rdflib.plugins.sparql.processor
from rdflib import Namespace, Literal, URIRef

from . import graph

GraphTripleComponent = Optional[Union[str, Namespace, Literal, URIRef]]
GraphTripleType = Tuple[Optional[GraphTripleComponent], Optional[GraphTripleComponent], Optional[GraphTripleComponent]]

def query(triple: GraphTripleType) -> List[GraphTripleType]:
    triple = transform_triple(triple)
    return list(map(lambda x: x, g().triples(triple)))

def sparql(sparql_q: str) -> rdflib.plugins.sparql.processor.SPARQLResult:
    return g().query(sparql_q)

#
# Helper Functions
#

def transform_triple(triple: GraphTripleType):
    s, p, o = triple
    return to_query_form(s), to_query_form(p), to_query_form(o)

def to_query_form(component: GraphTripleComponent):
    if not component or isinstance(component, URIRef) or isinstance(component, Literal):
        return component
    if isinstance(component, str):
        return Literal(component)
    return URIRef(component)

def g():
    return graph.graph()


from .rdf_prefix import *

def bind(g):
    g.bind('skos', skos)
    g.bind('owl', owl)
    g.bind('fibo-fnd-acc-cur', fibo_fnd_acc_cur)
    g.bind('fibo-fnd-utl-av',  fibo_fnd_utl_av)
    g.bind('fibo-fnd-acc-4217', fibo_fnd_acc_4217)
    g.bind('lcc-3166-1', lcc_3166_1)
    g.bind('lcc-cr', lcc_cr)
    g.bind('lcc-lr', lcc_lr)
    g.bind('sfo-cmn-ind-cur', sfo_cmn_ind_cur)
    return g



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rdf_cty_ccy',
 'rdf_cty_ccy.common',
 'rdf_cty_ccy.graph',
 'rdf_cty_ccy.model',
 'rdf_cty_ccy.query',
 'rdf_cty_ccy.rdfdata']

package_data = \
{'': ['*']}

install_requires = \
['PyMonad>=2.4.0,<3.0.0',
 'rdflib>=6.1.1,<7.0.0',
 'requests>=2.28.1,<3.0.0',
 'simple-memory-cache>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['build_onto = rdf_cty_ccy.common.build_onto:builder']}

setup_kwargs = {
    'name': 'rdf-cty-ccy',
    'version': '0.1.7',
    'description': 'Provides the FIBO and OMG Country and Currency Ontology in a queryable form.',
    'long_description': '# RDF Country and Currency Graph Library\n\nThis library builds an RDF graph of country and currency codes based on the following ontologies:\n\n+ https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/\n+ https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/\n\nBoth of these graphs are in Turtle format and are used extensively in the [FIBO ontologies](https://github.com/edmcouncil/fibo)\n\nThe Library is designed to support querying these triples in 3 modes:\n\n+ Using the Python RDFLib library.  The graph is queried using the RDFLib triples mode.  Returns RDF triples.\n+ Using a SPARQL query.  Returns an RDFLib SPARQL result.\n+ Via an OO query approach, which returns country and currency objects.\n\n# Building\n\nThe TTL files are extracted from the ontology locations (FIBO and OMG) and written to `rdf_cty_ccy/rdfdata`.  To re-get and re-build the ttl files, run the Makefile.\n\nThe Country and Currency ontologies are sourced from the following locations:\n\n+ [Currency Codes](https://spec.edmcouncil.org/fibo/ontology/master/2022Q1/FND/Accounting/ISO4217-CurrencyCodes/)\n+ [Country Codes](https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/)\n\n\n```shell\nmake build_cty_ccy_onto\n```\n\n\n# Usage\n\n## RDFLib Triples Mode\n\nThe triples are read into an inmemory RDFLib graph.  Therefore, using the rdflib.triples query format is supported.\n\nThe module `rdf_cty_ccy.graph.rdf_prefix` provides shorthand prefixes for the various RDF prefixes using in the ontologies.\n\n```python\nfibo_fnd_acc_cur = Namespace(\'https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/CurrencyAmount/\')\nfibo_fnd_acc_4217 = Namespace(\'https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/\')\nfibo_fnd_utl_av = Namespace("https://spec.edmcouncil.org/fibo/ontology/FND/Utilities/AnnotationVocabulary/")\nlcc_3166_1 = Namespace(\'https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/\')\nlcc_cr = Namespace(\'https://www.omg.org/spec/LCC/Countries/CountryRepresentation/\')\nlcc_lr = Namespace("https://www.omg.org/spec/LCC/Languages/LanguageRepresentation/")\n```\n\nTherefore `rdf_prefix.lcc_3166_1.NZL` is equivalent to the URI `https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NZL`\n\nThe RDFLib triples query produces a list of RDF triples which match the triples provided.  (See the docs at rdflib)[https://rdflib.readthedocs.io/en/stable/intro_to_graphs.html#basic-triple-matching].\n\n```python\nfrom rdf_cty_ccy.graph import graph, graph_query\nfrom rdf_cty_ccy.graph import rdf_prefix as P\n\nresult = graph_query.query((None, P.lcc_lr.hasTag, Literal("NZL", datatype=XSD.string)))\n\ns, _, _ = result[0]\n\ns == P.lcc_3166_1.NZL\n```\n\n\n## SPARQL Mode\n\n## Python Objects Mode\n\nThe graph can be queried using a python API interface; from the module `rdf_cty_ccy.query.query`.  It returns a `Country` object which has the following properties (note that `URIRef` and `Literal` comes from RDFLib:\n\n+ `country_uri`: URIRef \n+ `identifies`: URIRef\n+ `label`: Literal\n+ `currency`: Currency:\n  + `currency_uri`: URIRef\n  + `identifies`: URIRef\n  + `label`: Literal\n  \n\n```python\nfrom rdf_cty_ccy.query import query as Q\n\ncountry = Q.by_country_code(code=\'NZL\')\n\ncountry.country_uri             # => rdflib.term.URIRef(\'https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NZL\')\nstr(country.country_uri)        # => \'https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NZL\'\n```\n\nProvide the currency filter to obtain the currency properties for the country.\n\n```python\ncountry = Q.by_country_code(code=\'NZL\', filters=[Q.Filter.WithCurrency])\n\ncountry.currency.currency_uri   # => rdflib.term.URIRef(\'https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/NZD\')\n\n```\n\nYou can also query by the country URI (as a string) as follow:\n\n```python\ncountry = Q.by_country_uri(uri=\'https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NZL\')\n```\n\n',
    'author': 'Col Perks',
    'author_email': 'wild.fauve@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/wildfauve/rdf_cty_ccy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from typing import List
from dataclasses import dataclass, field
from rdflib import URIRef, Literal

@dataclass
class Currency:
    """
    <https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/NZD>
      a owl:NamedIndividual, fibo-fnd-acc-cur:CurrencyIdentifier ;
      rdfs:label "NZD" ;
      skos:definition "the currency identifier for New Zealand Dollar"@en ;
      lcc-lr:denotes <https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/NewZealandDollar> ;
      lcc-lr:hasTag "NZD" ;
      lcc-lr:identifies <https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/NewZealandDollar> ;
      lcc-lr:isMemberOf <https://spec.edmcouncil.org/fibo/ontology/FND/Accounting/ISO4217-CurrencyCodes/ISO4217-CodeSet> .
    """
    currency_uri: URIRef
    identifies: URIRef
    label: Literal


@dataclass
class Country:
    """
    <https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NZL>
      a owl:NamedIndividual, lcc-cr:Alpha3Code ;
      skos:definition "Alpha-3 country code for New Zealand" ;
      rdfs:label "NZL" ;
      lcc-lr:hasTag "NZL"^^xsd:string ;
      lcc-lr:denotes <https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NewZealand> ;
      lcc-lr:identifies <https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/NewZealand> ;
      lcc-lr:isMemberOf <https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/ISO3166-1-Alpha3-CodeSet> ;
      rdfs:isDefinedBy <https://www.omg.org/spec/LCC/Countries/ISO3166-1-CountryCodes/> .
    """
    country_uri: URIRef = None
    identifies: URIRef = None
    label: Literal = None
    code_type: URIRef = None
    currency: Currency = None
    currencies: List[Currency] = field(default_factory=list)

    def has_single_currency(self):
        return len(self.currencies) == 1
